from contextlib import closing
import psycopg


class PostgreSQLConnectionChecker:
    def __init__(self, **kwargs):
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.user = kwargs.get('user')
        self.password = kwargs.get('password')
        self.database = kwargs.get('database', 'postgres')

    def _get_connection(self):
        conn = psycopg.connect(f'host={self.host} port={self.port} dbname={self.database} user={self.user} password={self.password}', connect_timeout=10)
        return conn

    def check_connection(self):
        try:
            con = self._get_connection()
            with closing(con) as con:
                cur = con.cursor()
                cur.execute('select 1;')
            connected = True
        except Exception as e:
            print('EXCEPTION!')
            print(e)
            connected = False
        return connected



import os

from cassandra.cluster import Cluster
from cassandra.auth import PlainTextAuthProvider


class CassandraConnectionChecker:
    def __init__(self, **kwargs):
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.user = kwargs.get('user')
        self.password = kwargs.get('password')
        self.keyspace = kwargs.get('database')
        self.secure_connect_bundle = kwargs.get('secure_connect_bundle')
        self.protocol_version = kwargs.get('protocol_version')

    def check_connection(self):
        try:
            auth_provider = PlainTextAuthProvider(
                username=self.user, password=self.password
            )
            connection_props = {
                'auth_provider': auth_provider
            }

            if self.protocol_version is not None:
                connection_props['protocol_version'] = self.protocol_version

            if self.secure_connect_bundle is not None:
                if os.path.isfile(self.secure_connect_bundle) is False:
                    raise Exception("'secure_connect_bundle' must be path to the file")
                connection_props['cloud'] = {
                    'secure_connect_bundle': self.secure_connect_bundle
                }
            else:
                connection_props['contact_points'] = [self.host]
                connection_props['port'] = self.port

            cluster = Cluster(**connection_props)
            session = cluster.connect()

            if isinstance(self.keyspace, str) and len(self.keyspace) > 0:
                session.set_keyspace(self.keyspace)

            session.execute('SELECT release_version FROM system.local;').one()

            connected = True
        except Exception:
            connected = False
        return connected



from functools import partial

import pandas as pd

from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import get_all_tables
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.file_datanode import FileDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.view_datanode import ViewDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.mindsdb_datanode import MindsDBDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.integration_datanode import IntegrationDataNode


class InformationSchema(DataNode):
    type = 'INFORMATION_SCHEMA'

    information_schema = {
        'SCHEMATA': ['CATALOG_NAME', 'SCHEMA_NAME', 'DEFAULT_CHARACTER_SET_NAME', 'DEFAULT_COLLATION_NAME', 'SQL_PATH'],
        'TABLES': ['TABLE_NAME', 'TABLE_SCHEMA', 'TABLE_TYPE', 'TABLE_ROWS', 'TABLE_COLLATION', 'TABLE_COMMENT'],
        'COLUMNS': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'COLUMN_NAME', 'ORDINAL_POSITION', 'COLUMN_DEFAULT', 'IS_NULLABLE', 'DATA_TYPE', 'CHARACTER_MAXIMUM_LENGTH', 'CHARACTER_OCTET_LENGTH', 'NUMERIC_PRECISION', 'NUMERIC_SCALE', 'DATETIME_PRECISION', 'CHARACTER_SET_NAME', 'COLLATION_NAME', 'COLUMN_TYPE', 'COLUMN_KEY', 'EXTRA', 'PRIVILEGES', 'COLUMN_COMMENT', 'GENERATION_EXPRESSION'],
        'EVENTS': ['EVENT_CATALOG', 'EVENT_SCHEMA', 'EVENT_NAME', 'DEFINER', 'TIME_ZONE', 'EVENT_BODY', 'EVENT_DEFINITION', 'EVENT_TYPE', 'EXECUTE_AT', 'INTERVAL_VALUE', 'INTERVAL_FIELD', 'SQL_MODE', 'STARTS', 'ENDS', 'STATUS', 'ON_COMPLETION', 'CREATED', 'LAST_ALTERED', 'LAST_EXECUTED', 'EVENT_COMMENT', 'ORIGINATOR', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'ROUTINES': ['SPECIFIC_NAME', 'ROUTINE_CATALOG', 'ROUTINE_SCHEMA', 'ROUTINE_NAME', 'ROUTINE_TYPE', 'DATA_TYPE', 'CHARACTER_MAXIMUM_LENGTH', 'CHARACTER_OCTET_LENGTH', 'NUMERIC_PRECISION', 'NUMERIC_SCALE', 'DATETIME_PRECISION', 'CHARACTER_SET_NAME', 'COLLATION_NAME', 'DTD_IDENTIFIER', 'ROUTINE_BODY', 'ROUTINE_DEFINITION', 'EXTERNAL_NAME', 'EXTERNAL_LANGUAGE', 'PARAMETER_STYLE', 'IS_DETERMINISTIC', 'SQL_DATA_ACCESS', 'SQL_PATH', 'SECURITY_TYPE', 'CREATED', 'LAST_ALTERED', 'SQL_MODE', 'ROUTINE_COMMENT', 'DEFINER', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'TRIGGERS': ['TRIGGER_CATALOG', 'TRIGGER_SCHEMA', 'TRIGGER_NAME', 'EVENT_MANIPULATION', 'EVENT_OBJECT_CATALOG', 'EVENT_OBJECT_SCHEMA', 'EVENT_OBJECT_TABLE', 'ACTION_ORDER', 'ACTION_CONDITION', 'ACTION_STATEMENT', 'ACTION_ORIENTATION', 'ACTION_TIMING', 'ACTION_REFERENCE_OLD_TABLE', 'ACTION_REFERENCE_NEW_TABLE', 'ACTION_REFERENCE_OLD_ROW', 'ACTION_REFERENCE_NEW_ROW', 'CREATED', 'SQL_MODE', 'DEFINER', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'PLUGINS': ['PLUGIN_NAME', 'PLUGIN_VERSION', 'PLUGIN_STATUS', 'PLUGIN_TYPE', 'PLUGIN_TYPE_VERSION', 'PLUGIN_LIBRARY', 'PLUGIN_LIBRARY_VERSION', 'PLUGIN_AUTHOR', 'PLUGIN_DESCRIPTION', 'PLUGIN_LICENSE', 'LOAD_OPTION', 'PLUGIN_MATURITY', 'PLUGIN_AUTH_VERSION'],
        'ENGINES': ['ENGINE', 'SUPPORT', 'COMMENT', 'TRANSACTIONS', 'XA', 'SAVEPOINTS'],
        'KEY_COLUMN_USAGE': ['CONSTRAINT_CATALOG', 'CONSTRAINT_SCHEMA', 'CONSTRAINT_NAME', 'TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'COLUMN_NAME', 'ORDINAL_POSITION', 'POSITION_IN_UNIQUE_CONSTRAINT', 'REFERENCED_TABLE_SCHEMA', 'REFERENCED_TABLE_NAME', 'REFERENCED_COLUMN_NAME'],
        'STATISTICS': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'NON_UNIQUE', 'INDEX_SCHEMA', 'INDEX_NAME', 'SEQ_IN_INDEX', 'COLUMN_NAME', 'COLLATION', 'CARDINALITY', 'SUB_PART', 'PACKED', 'NULLABLE', 'INDEX_TYPE', 'COMMENT', 'INDEX_COMMENT', 'IS_VISIBLE', 'EXPRESSION'],
        'CHARACTER_SETS': ['CHARACTER_SET_NAME', 'DEFAULT_COLLATE_NAME', 'DESCRIPTION', 'MAXLEN'],
        'COLLATIONS': ['COLLATION_NAME', 'CHARACTER_SET_NAME', 'ID', 'IS_DEFAULT', 'IS_COMPILED', 'SORTLEN', 'PAD_ATTRIBUTE'],
    }

    def __init__(self, session):
        self.session = session
        self.integration_controller = session.integration_controller
        self.data_store = session.data_store
        self.view_interface = session.view_interface
        self.persis_datanodes = {
            'mindsdb': MindsDBDataNode(
                session.model_interface,
                session.data_store,
                session.integration_controller
            ),
            'files': FileDataNode(session.data_store),
            'views': ViewDataNode(session.view_interface, session.integration_controller, session.data_store)
        }

        self.get_dataframe_funcs = {
            'TABLES': self._get_tables,
            'COLUMNS': self._get_columns,
            'SCHEMATA': self._get_schemata,
            'ENGINES': self._get_engines,
            'CHARACTER_SETS': self._get_charsets,
            'COLLATIONS': self._get_collations,
        }
        for table_name in self.information_schema:
            if table_name not in self.get_dataframe_funcs:
                self.get_dataframe_funcs[table_name] = partial(self._get_empty_table, table_name)

    def __getitem__(self, key):
        return self.get(key)

    def get(self, name):
        name_lower = name.lower()

        if name.lower() == 'information_schema':
            return self

        if name_lower in self.persis_datanodes:
            return self.persis_datanodes[name_lower]

        datasource_names = self.integration_controller.get_all().keys()
        for datasource_name in datasource_names:
            if datasource_name.lower() == name_lower:
                datasource = self.integration_controller.get(name=datasource_name)
                return IntegrationDataNode(datasource_name, self.data_store, ds_type=datasource['type'])

        return None

    def has_table(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return True
        return False

    def get_table_columns(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return self.information_schema[tn]
        raise Exception(f'Table information_schema.{tableName} does not exists')

    def get_integrations_names(self):
        integration_names = self.integration_controller.get_all().keys()
        return [
            x.lower() for x in integration_names
        ]

    def _get_tables(self):
        columns = self.information_schema['TABLES']

        data = [
            [name, 'information_schema', 'SYSTEM VIEW', [], 'utf8mb4_0900_ai_ci', None]
            for name in self.information_schema.keys()
        ]

        for ds_name, ds in self.persis_datanodes.items():
            ds_tables = ds.get_tables()
            data += [[x, ds_name, 'BASE TABLE', [], 'utf8mb4_0900_ai_ci', None] for x in ds_tables]

        for ds_name in self.get_integrations_names():
            ds = self.get(ds_name)
            ds_tables = ds.get_tables()
            data += [[x, ds_name, 'BASE TABLE', [], 'utf8mb4_0900_ai_ci', None] for x in ds_tables]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_columns(self):
        columns = self.information_schema['COLUMNS']

        # NOTE there is a lot of types in mysql, but listed below should be enough for our purposes
        row_templates = {
            'text': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'varchar', 1024, 3072, None, None, None, 'utf8', 'utf8_bin', 'varchar(1024)', None, None, 'select', None, None],
            'timestamp': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', 'CURRENT_TIMESTAMP', 'YES', 'timestamp', None, None, None, None, 0, None, None, 'timestamp', None, None, 'select', None, None],
            'bigint': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'bigint', None, None, 20, 0, None, None, None, 'bigint unsigned', None, None, 'select', None, None],
            'float': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'float', None, None, 12, 0, None, None, None, 'float', None, None, 'select', None, None]
        }

        result = []

        for table_name in self.information_schema:
            table_columns = self.information_schema[table_name]
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'information_schema'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        mindsb_dn = self.get('MINDSDB')
        for table_name in mindsb_dn.get_tables():
            table_columns = mindsb_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'mindsdb'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        mindsb_dn = self.get('FILES')
        for table_name in mindsb_dn.get_tables():
            table_columns = mindsb_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'files'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        df = pd.DataFrame(result, columns=columns)
        return df

    def _get_schemata(self):
        columns = self.information_schema['SCHEMATA']
        data = [
            ['def', 'information_schema', 'utf8', 'utf8_general_ci', None]
        ]

        for database_name in self.persis_datanodes:
            data.append(['def', database_name, 'utf8mb4', 'utf8mb4_0900_ai_ci', None])

        integration_names = self.integration_controller.get_all().keys()
        for database_name in integration_names:
            data.append(['def', database_name, 'utf8mb4', 'utf8mb4_0900_ai_ci', None])

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_engines(self):
        columns = self.information_schema['ENGINES']
        data = [['InnoDB', 'DEFAULT', 'Supports transactions, row-level locking, and foreign keys', 'YES', 'YES', 'YES']]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_charsets(self):
        columns = self.information_schema['CHARACTER_SETS']
        data = [
            ['utf8', 'UTF-8 Unicode', 'utf8_general_ci', 3],
            ['latin1', 'cp1252 West European', 'latin1_swedish_ci', 1],
            ['utf8mb4', 'UTF-8 Unicode', 'utf8mb4_general_ci', 4]
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_collations(self):
        columns = self.information_schema['COLLATIONS']
        data = [
            ['utf8_general_ci', 'utf8', 33, 'Yes', 'Yes', 1, 'PAD SPACE'],
            ['latin1_swedish_ci', 'latin1', 8, 'Yes', 'Yes', 1, 'PAD SPACE']
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_empty_table(self, table_name):
        columns = self.information_schema[table_name]
        data = []

        df = pd.DataFrame(data, columns=columns)
        return df

    def select(self, query):
        query_tables = get_all_tables(query)

        if len(query_tables) != 1:
            raise Exception(f'Only one table can be used in query to information_schema: {query}')

        table_name = query_tables[0].upper()

        if table_name not in self.get_dataframe_funcs:
            raise Exception('Information schema: Not implemented.')

        dataframe = self.get_dataframe_funcs[table_name]()

        try:
            data = query_df(dataframe, query, session=self.session)
        except Exception as e:
            print(f'Exception! {e}')
            return [], []

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in data.dtypes.items()
        ]

        return data.to_dict(orient='records'), columns_info



from contextlib import closing
import mysql.connector


class MySQLConnectionChecker:
    def __init__(self, **kwargs):
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.user = kwargs.get('user')
        self.password = kwargs.get('password')
        self.ssl = kwargs.get('ssl')
        self.ssl_ca = kwargs.get('ssl_ca')
        self.ssl_cert = kwargs.get('ssl_cert')
        self.ssl_key = kwargs.get('ssl_key')

    def _get_connnection(self):
        config = {
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password
        }
        if self.ssl is True:
            config['client_flags'] = [mysql.connector.constants.ClientFlag.SSL]
            if self.ssl_ca is not None:
                config["ssl_ca"] = self.ssl_ca
            if self.ssl_cert is not None:
                config["ssl_cert"] = self.ssl_cert
            if self.ssl_key is not None:
                config["ssl_key"] = self.ssl_key
        return mysql.connector.connect(**config)

    def check_connection(self):
        try:
            con = self._get_connnection()
            with closing(con) as con:
                connected = con.is_connected()
        except Exception:
            connected = False
        return connected



import requests


class ClickhouseConnectionChecker:
    def __init__(self, **kwargs):
        self.host = kwargs.get("host")
        self.port = kwargs.get("port")
        self.user = kwargs.get("user")
        self.password = kwargs.get("password")

    def check_connection(self):
        try:
            res = requests.post(f"http://{self.host}:{self.port}",
                                data="select 1;",
                                params={'user': self.user, 'password': self.password})
            connected = res.status_code == 200
        except Exception:
            connected = False
        return connected



import os
import shutil
import tempfile
from pathlib import Path
from copy import deepcopy

from sqlalchemy import func

from mindsdb.interfaces.storage.db import session, Integration
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.utilities.fs import create_directory
from mindsdb.integrations import CHECKERS as DB_CONNECTION_CHECKERS


class IntegrationController:
    @staticmethod
    def _is_not_empty_str(s):
        return isinstance(s, str) and len(s) > 0

    def add(self, name, data, company_id=None):
        if 'database_name' not in data:
            data['database_name'] = name
        if 'publish' not in data:
            data['publish'] = True

        bundle_path = data.get('secure_connect_bundle')
        if data.get('type') in ('cassandra', 'scylla') and self._is_not_empty_str(bundle_path):
            if os.path.isfile(bundle_path) is False:
                raise Exception(f'Can not get access to file: {bundle_path}')
            integrations_dir = Config()['paths']['integrations']

            p = Path(bundle_path)
            data['secure_connect_bundle'] = p.name

            integration_record = Integration(name=name, data=data, company_id=company_id)
            session.add(integration_record)
            session.commit()
            integration_id = integration_record.id

            folder_name = f'integration_files_{company_id}_{integration_id}'
            integration_dir = os.path.join(integrations_dir, folder_name)
            create_directory(integration_dir)
            shutil.copyfile(bundle_path, os.path.join(integration_dir, p.name))

            FsStore().put(
                folder_name,
                integration_dir,
                integrations_dir
            )
        elif data.get('type') in ('mysql', 'mariadb'):
            ssl = data.get('ssl')
            files = {}
            temp_dir = None
            if ssl is True:
                for key in ['ssl_ca', 'ssl_cert', 'ssl_key']:
                    if key not in data:
                        continue
                    if os.path.isfile(data[key]) is False:
                        if self._is_not_empty_str(data[key]) is False:
                            raise Exception("'ssl_ca', 'ssl_cert' and 'ssl_key' must be paths or inline certs")
                        if temp_dir is None:
                            temp_dir = tempfile.mkdtemp(prefix='integration_files_')
                        cert_file_name = data.get(f'{key}_name', f'{key}.pem')
                        cert_file_path = os.path.join(temp_dir, cert_file_name)
                        with open(cert_file_path, 'wt') as f:
                            f.write(data[key])
                        data[key] = cert_file_path
                    files[key] = data[key]
                    p = Path(data[key])
                    data[key] = p.name
            integration_record = Integration(name=name, data=data, company_id=company_id)
            session.add(integration_record)
            session.commit()
            integration_id = integration_record.id

            if len(files) > 0:
                integrations_dir = Config()['paths']['integrations']
                folder_name = f'integration_files_{company_id}_{integration_id}'
                integration_dir = os.path.join(integrations_dir, folder_name)
                create_directory(integration_dir)
                for file_path in files.values():
                    p = Path(file_path)
                    shutil.copyfile(file_path, os.path.join(integration_dir, p.name))
                FsStore().put(
                    folder_name,
                    integration_dir,
                    integrations_dir
                )
        else:
            integration_record = Integration(name=name, data=data, company_id=company_id)
            session.add(integration_record)
            session.commit()

    def modify(self, name, data, company_id):
        integration_record = session.query(Integration).filter_by(company_id=company_id, name=name).first()
        old_data = deepcopy(integration_record.data)
        for k in old_data:
            if k not in data:
                data[k] = old_data[k]

        integration_record.data = data
        session.commit()

    def delete(self, name, company_id=None):
        integration_record = session.query(Integration).filter_by(company_id=company_id, name=name).first()
        integrations_dir = Config()['paths']['integrations']
        folder_name = f'integration_files_{company_id}_{integration_record.id}'
        integration_dir = os.path.join(integrations_dir, folder_name)
        if os.path.isdir(integration_dir):
            shutil.rmtree(integration_dir)
        try:
            FsStore().delete(folder_name)
        except Exception:
            pass
        session.delete(integration_record)
        session.commit()

    def _get_integration_record_data(self, integration_record, sensitive_info=True):
        if integration_record is None or integration_record.data is None:
            return None
        data = deepcopy(integration_record.data)
        if data.get('password', None) is None:
            data['password'] = ''
        data['date_last_update'] = deepcopy(integration_record.updated_at)

        bundle_path = data.get('secure_connect_bundle')
        mysql_ssl_ca = data.get('ssl_ca')
        mysql_ssl_cert = data.get('ssl_cert')
        mysql_ssl_key = data.get('ssl_key')
        if (
            data.get('type') in ('mysql', 'mariadb')
            and (
                self._is_not_empty_str(mysql_ssl_ca)
                or self._is_not_empty_str(mysql_ssl_cert)
                or self._is_not_empty_str(mysql_ssl_key)
            )
            or data.get('type') in ('cassandra', 'scylla')
            and bundle_path is not None
        ):
            fs_store = FsStore()
            integrations_dir = Config()['paths']['integrations']
            folder_name = f'integration_files_{integration_record.company_id}_{integration_record.id}'
            integration_dir = os.path.join(integrations_dir, folder_name)
            fs_store.get(
                folder_name,
                integration_dir,
                integrations_dir
            )

        if not sensitive_info:
            if 'password' in data:
                data['password'] = None
            if (
                data.get('type') == 'redis'
                and isinstance(data.get('connection'), dict)
                and 'password' in data['connection']
            ):
                data['connection'] = None

        data['id'] = integration_record.id
        data['name'] = integration_record.name

        return data

    def get_by_id(self, id, company_id=None, sensitive_info=True):
        integration_record = session.query(Integration).filter_by(company_id=company_id, id=id).first()
        return self._get_integration_record_data(integration_record, sensitive_info)

    def get(self, name, company_id=None, sensitive_info=True, case_sensitive=False):
        if case_sensitive:
            integration_record = session.query(Integration).filter_by(company_id=company_id, name=name).first()
        else:
            integration_record = session.query(Integration).filter(
                (Integration.company_id == company_id)
                & (func.lower(Integration.name) == func.lower(name))
            ).first()
        return self._get_integration_record_data(integration_record, sensitive_info)

    def get_all(self, company_id=None, sensitive_info=True):
        integration_records = session.query(Integration).filter_by(company_id=company_id).all()
        integration_dict = {}
        for record in integration_records:
            if record is None or record.data is None:
                continue
            integration_dict[record.name] = self._get_integration_record_data(record, sensitive_info)
        return integration_dict

    def check_connections(self):
        connections = {}
        for integration_name, integration_meta in self.get_all().items():
            connection_checker = DB_CONNECTION_CHECKERS.get(integration_meta.get('type'))
            if connection_checker is not None:
                status = connection_checker(**integration_meta).check_connection()
                connections[integration_name] = status
            else:
                connections[integration_name] = True

        return connections



from .clickhouse.clickhouse import ClickhouseConnectionChecker
from .mariadb.mariadb import MariadbConnectionChecker
from .mongodb.mongodb import MongoConnectionChecker
from .mssql.mssql import MSSQLConnectionChecker
from .mysql.mysql import MySQLConnectionChecker
from .postgres.postgres import PostgreSQLConnectionChecker
from .redis.redisdb import RedisConnectionChecker
from .kafka.kafkadb import KafkaConnectionChecker
from .snowflake.snowflake import SnowflakeConnectionChecker
from .trinodb.trinodb import TrinodbConnectionChecker

try:
    from .scylladb.scylladb import ScyllaDBConnectionChecker
except ImportError:
    ScyllaDBConnectionChecker = None
try:
    from .cassandra.cassandra import CassandraConnectionChecker
except ImportError:
    CassandraConnectionChecker = None


CHECKERS = {
    "clickhouse": ClickhouseConnectionChecker,
    "mariadb": MariadbConnectionChecker,
    "mongodb": MongoConnectionChecker,
    "mssql": MSSQLConnectionChecker,
    "mysql": MySQLConnectionChecker,
    "singlestore": MySQLConnectionChecker,
    "postgres": PostgreSQLConnectionChecker,
    "cockroachdb": PostgreSQLConnectionChecker,
    "redis": RedisConnectionChecker,
    "kafka": KafkaConnectionChecker,
    "snowflake": SnowflakeConnectionChecker,
    "trinodb": TrinodbConnectionChecker
}


if ScyllaDBConnectionChecker is not None:
    CHECKERS['scylladb'] = ScyllaDBConnectionChecker

if CassandraConnectionChecker is not None:
    CHECKERS['cassandra'] = CassandraConnectionChecker



from mindsdb.integrations.postgres_handler.postgres_handler import PostgresHandler

class QuestDBHandler(PostgresHandler):
    """
    This handler handles connection and execution of the QuestDB statements. 
    TODO: check the dialect for questdb
    """

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    def get_tables(self):
        """
        List all tabels in QuestDB
        """
        query = "SHOW TABLES"
        res = super().native_query(query)
        return res

    def describe_table(self, table_name):
        """
        List information about the table
        """
        query = f"SELECT * FROM tables() WHERE name='{table_name}';"
        result = super().native_query(query)
        return result
    
    def get_views(self):
        """
        QuestDB doesn't support views
        """
        raise NotImplementedError()



import copy
import datetime
from dateutil.parser import parse as parse_datetime
import subprocess
import os
import sys
import tempfile
import shutil
from pathlib import Path

from flask import request
from flask_restx import Resource, abort
from flask import current_app as ca
from dateutil.tz import tzlocal

from mindsdb.utilities.log import log
from mindsdb.api.http.namespaces.configs.config import ns_conf
from mindsdb.utilities.log import get_logs
from mindsdb.integrations import CHECKERS
from mindsdb.api.http.utils import http_error
from mindsdb.interfaces.stream.stream import StreamController


@ns_conf.route('/logs')
@ns_conf.param('name', 'Get logs')
class GetLogs(Resource):
    @ns_conf.doc('get_integrations')
    def get(self):
        min_timestamp = parse_datetime(request.args['min_timestamp'])
        max_timestamp = request.args.get('max_timestamp', None)
        context = request.args.get('context', None)
        level = request.args.get('level', None)
        log_from = request.args.get('log_from', None)
        limit = request.args.get('limit', None)

        logs = get_logs(min_timestamp, max_timestamp, context, level, log_from, limit)
        return {'data': logs}


@ns_conf.route('/integrations')
@ns_conf.param('name', 'List all database integration')
class ListIntegration(Resource):
    def get(self):
        return {
            'integrations': [k for k in request.integration_controller.get_all(sensitive_info=False)]
        }


@ns_conf.route('/all_integrations')
@ns_conf.param('name', 'List all database integration')
class AllIntegration(Resource):
    @ns_conf.doc('get_all_integrations')
    def get(self):
        integrations = request.integration_controller.get_all(sensitive_info=False)
        return integrations


@ns_conf.route('/integrations/<name>')
@ns_conf.param('name', 'Database integration')
class Integration(Resource):
    @ns_conf.doc('get_integration')
    def get(self, name):
        integration = request.integration_controller.get(name, sensitive_info=False)
        if integration is None:
            abort(404, f'Can\'t find database integration: {name}')
        integration = copy.deepcopy(integration)
        return integration

    @ns_conf.doc('put_integration')
    def put(self, name):
        params = {}
        params.update((request.json or {}).get('params', {}))
        params.update(request.form or {})

        if len(params) == 0:
            abort(400, "type of 'params' must be dict")

        # params from FormData will be as text
        for key in ('publish', 'test', 'enabled'):
            if key in params:
                if isinstance(params[key], str) and params[key].lower() in ('false', '0'):
                    params[key] = False
                else:
                    params[key] = bool(params[key])

        files = request.files
        temp_dir = None
        if files is not None:
            temp_dir = tempfile.mkdtemp(prefix='integration_files_')
            for key, file in files.items():
                temp_dir_path = Path(temp_dir)
                file_name = Path(file.filename)
                file_path = temp_dir_path.joinpath(file_name).resolve()
                if temp_dir_path not in file_path.parents:
                    raise Exception(f'Can not save file at path: {file_path}')
                file.save(file_path)
                params[key] = file_path

        is_test = params.get('test', False)
        if is_test:
            del params['test']
            db_type = params.get('type')
            checker_class = CHECKERS.get(db_type, None)
            if checker_class is None:
                abort(400, f"Unknown integration type: {db_type}")
            checker = checker_class(**params)
            if temp_dir is not None:
                shutil.rmtree(temp_dir)
            return {'success': checker.check_connection()}, 200

        integration = request.integration_controller.get(name, sensitive_info=False)
        if integration is not None:
            abort(400, f"Integration with name '{name}' already exists")

        try:
            if 'enabled' in params:
                params['publish'] = params['enabled']
                del params['enabled']
            request.integration_controller.add(name, params)

            model_data_arr = []
            for model in request.model_interface.get_models():
                if model['status'] == 'complete':
                    try:
                        model_data_arr.append(request.model_interface.get_model_data(model['name']))
                    except Exception:
                        pass

            if is_test is False and params.get('publish', False) is True:
                model_data_arr = []
                for model in request.model_interface.get_models():
                    if model['status'] == 'complete':
                        try:
                            model_data_arr.append(request.model_interface.get_model_data(model['name']))
                        except Exception:
                            pass

                stream_controller = StreamController(request.company_id)
                if params.get('type') in stream_controller.known_dbs and params.get('publish', False) is True:
                    stream_controller.setup(name)
        except Exception as e:
            log.error(str(e))
            if temp_dir is not None:
                shutil.rmtree(temp_dir)
            abort(500, f'Error during config update: {str(e)}')

        if temp_dir is not None:
            shutil.rmtree(temp_dir)
        return '', 200

    @ns_conf.doc('delete_integration')
    def delete(self, name):
        integration = request.integration_controller.get(name)
        if integration is None:
            abort(400, f"Nothing to delete. '{name}' not exists.")
        try:
            request.integration_controller.delete(name)
        except Exception as e:
            log.error(str(e))
            abort(500, f'Error during integration delete: {str(e)}')
        return '', 200

    @ns_conf.doc('modify_integration')
    def post(self, name):
        params = {}
        params.update((request.json or {}).get('params', {}))
        params.update(request.form or {})

        if not isinstance(params, dict):
            abort(400, "type of 'params' must be dict")
        integration = request.integration_controller.get(name)
        if integration is None:
            abort(400, f"Nothin to modify. '{name}' not exists.")
        try:
            if 'enabled' in params:
                params['publish'] = params['enabled']
                del params['enabled']
            request.integration_controller.modify(name, params)

            stream_controller = StreamController(request.company_id)
            if params.get('type') in stream_controller.known_dbs and params.get('publish', False) is True:
                stream_controller.setup(name)
        except Exception as e:
            log.error(str(e))
            abort(500, f'Error during integration modifycation: {str(e)}')
        return '', 200


@ns_conf.route('/integrations/<name>/check')
@ns_conf.param('name', 'Database integration checks')
class Check(Resource):
    @ns_conf.doc('check')
    def get(self, name):
        if request.integration_controller.get(name) is None:
            abort(404, f'Can\'t find database integration: {name}')
        connections = request.integration_controller.check_connections()
        return connections.get(name, False), 200


@ns_conf.route('/vars')
class Vars(Resource):
    def get(self):
        if os.getenv('CHECK_FOR_UPDATES', '1').lower() in ['0', 'false']:
            telemtry = False
        else:
            telemtry = True

        if ca.config_obj.get('disable_mongo', False):
            mongo = False
        else:
            mongo = True

        cloud = ca.config_obj.get('cloud', False)
        local_time = datetime.datetime.now(tzlocal())
        local_timezone = local_time.tzname()

        return {
            'mongo': mongo,
            'telemtry': telemtry,
            'cloud': cloud,
            'timezone': local_timezone,
        }


@ns_conf.route('/install_options')
@ns_conf.param('dependency_list', 'Install dependencies')
class InstallDependenciesList(Resource):
    def get(self):
        return {'dependencies': ['snowflake', 'athena', 'google', 's3', 'lightgbm_gpu', 'mssql', 'cassandra', 'scylladb']}


@ns_conf.route('/install/<dependency>')
@ns_conf.param('dependency', 'Install dependencies')
class InstallDependencies(Resource):
    def get(self, dependency):
        if dependency == 'snowflake':
            dependency = ['snowflake-connector-python[pandas]', 'asn1crypto==1.3.0']
        elif dependency == 'athena':
            dependency = ['PyAthena >= 2.0.0']
        elif dependency == 'google':
            dependency = ['google-cloud-storage', 'google-auth']
        elif dependency == 's3':
            dependency = ['boto3 >= 1.9.0']
        elif dependency == 'lightgbm_gpu':
            dependency = ['lightgbm', '--install-option=--gpu', '--upgrade']
        elif dependency == 'mssql':
            dependency = ['pymssql >= 2.1.4']
        elif dependency == 'cassandra':
            dependency = ['cassandra-driver']
        elif dependency == 'scylladb':
            dependency = ['scylla-driver']
        else:
            return f'Unkown dependency: {dependency}', 400

        outs = b''
        errs = b''
        try:
            sp = subprocess.Popen(
                [sys.executable, '-m', 'pip', 'install', *dependency],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            code = sp.wait()
            outs, errs = sp.communicate(timeout=1)
        except Exception as e:
            return http_error(500, 'Failed to install dependency', str(e))

        if code != 0:
            output = ''
            if isinstance(outs, bytes) and len(outs) > 0:
                output = output + 'Output: ' + outs.decode()
            if isinstance(errs, bytes) and len(errs) > 0:
                if len(output) > 0:
                    output = output + '\n'
                output = output + 'Errors: ' + errs.decode()
            return http_error(500, 'Failed to install dependency', output)

        return 'Installed', 200



from mindsdb_datasources import FileDS

from mindsdb.api.mysql.mysql_proxy.classes.sql_query import get_all_tables
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df


class FileDataNode(DataNode):
    type = 'file'

    def __init__(self, data_store):
        self.data_store = data_store

    def get_tables(self):
        return self.data_store.get_files_names()

    def has_table(self, table):
        return table in self.get_tables()

    def get_table_columns(self, table):
        file_meta = self.data_store.get_file_meta(table)
        return [x['name'] for x in file_meta['columns']]

    def select(self, query):
        query_tables = get_all_tables(query)

        if len(query_tables) != 1:
            raise Exception(f'Only one table can be used in query to information_schema: {query}')

        file_path = self.data_store.get_file_path(query_tables[0])
        file_datasource = FileDS(file_path)
        data_df = file_datasource.df
        result = query_df(data_df, query)

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in result.dtypes.items()
        ]

        return result.to_dict(orient='records'), columns_info



from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import ERR

class SqlApiException(Exception):
    err_code = ERR.ER_SYNTAX_ERROR



class ErBadDbError(SqlApiException):
    err_code = ERR.ER_BAD_DB_ERROR

class ErBadTableError(SqlApiException):
    err_code = ERR.ER_BAD_DB_ERROR

class ErKeyColumnDoesNotExist(SqlApiException):
    err_code = ERR.ER_KEY_COLUMN_DOES_NOT_EXIST

class ErTableExistError(SqlApiException):
    err_code = ERR.ER_TABLE_EXISTS_ERROR

class ErDubFieldName(SqlApiException):
    err_code = ERR.ER_DUP_FIELDNAME

class ErDbDropDelete(SqlApiException):
    err_code = ERR.ER_DB_DROP_DELETE

class ErNonInsertableTable(SqlApiException):
    err_code = ERR.ER_NON_INSERTABLE_TABLE

class ErNotSupportedYet(SqlApiException):
    err_code = ERR.ER_NOT_SUPPORTED_YET



import json
from dateutil.parser import parse as parse_datetime
from flask import request
from flask_restx import Resource, abort
from mindsdb.utilities.log import log
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from mindsdb.api.http.namespaces.entitites.predictor_metadata import (
    predictor_query_params,
    put_predictor_params,
    put_predictor_metadata
)


@ns_conf.route('/')
class PredictorList(Resource):
    @ns_conf.doc('list_predictors')
    def get(self):
        '''List all predictors'''
        models = request.model_interface.get_models()
        return models


@ns_conf.route('/<name>')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class Predictor(Resource):
    @ns_conf.doc('get_predictor')
    def get(self, name):
        model = request.model_interface.get_model_data(name)

        for k in ['train_end_at', 'updated_at', 'created_at']:
            if k in model and model[k] is not None:
                model[k] = parse_datetime(model[k])

        return model

    @ns_conf.doc('delete_predictor')
    def delete(self, name):
        '''Remove predictor'''
        request.model_interface.delete_model(name)
        return '', 200

    @ns_conf.doc('put_predictor')
    @ns_conf.expect(put_predictor_metadata)
    def put(self, name):
        '''Learning new predictor'''
        data = request.json
        to_predict = data.get('to_predict')

        try:
            kwargs = data.get('kwargs')
        except Exception:
            kwargs = None

        if isinstance(kwargs, dict) is False:
            kwargs = {}

        if 'equal_accuracy_for_all_output_categories' not in kwargs:
            kwargs['equal_accuracy_for_all_output_categories'] = True

        if 'advanced_args' not in kwargs:
            kwargs['advanced_args'] = {}

        if 'use_selfaware_model' not in kwargs['advanced_args']:
            kwargs['advanced_args']['use_selfaware_model'] = False

        try:
            retrain = data.get('retrain')
            if retrain in ('true', 'True'):
                retrain = True
            else:
                retrain = False
        except Exception:
            retrain = None

        ds_name = data.get('data_source_name')
        from_ds = data.get('from')
        delete_ds_on_fail = False
        if ds_name is not None:
            ds = request.default_store.get_datasource_obj(ds_name, raw=True)
            if ds is None:
                return http_error(
                    400,
                    'DS not exists',
                    f'Can not find datasource: {ds_name}'
                )
        elif isinstance(from_ds, dict):
            if 'datasource' not in from_ds or 'query' not in from_ds:
                return http_error(
                    400,
                    'Wring arguments',
                    "'from' must contain 'datasource' and 'query'"
                )
            delete_ds_on_fail = True
            ds_name = request.default_store.get_vacant_name(name)

            if request.integration_controller.get(from_ds['datasource']) is None:
                return http_error(
                    400,
                    'Datasource not exist',
                    f"Datasource not exist: {from_ds['datasource']}"
                )

            ds = request.default_store.save_datasource(ds_name, from_ds['datasource'], {'query': from_ds['query']})
        else:
            return http_error(
                400,
                'Wring arguments',
                "query must contain 'data_source_name' or 'from'"
            )

        if retrain is True:
            original_name = name
            name = name + '_retrained'

        model_names = [x['name'] for x in request.model_interface.get_models()]
        if name in model_names:
            return http_error(
                409,
                f"Predictor '{name}' already exists",
                f"Predictor with name '{name}' already exists. Each predictor must have unique name."
            )

        request.model_interface.learn(
            name, ds, to_predict, request.default_store.get_datasource(ds_name)['id'],
            kwargs=kwargs, delete_ds_on_fail=delete_ds_on_fail, user_class=request.user_class
        )

        if retrain is True:
            try:
                request.model_interface.delete_model(original_name)
                request.model_interface.rename_model(name, original_name)
            except Exception:
                pass

        return '', 200


@ns_conf.route('/<name>/update')
@ns_conf.param('name', 'Update predictor')
class PredictorUpdate(Resource):
    @ns_conf.doc('Update predictor')
    def get(self, name):
        msg = request.model_interface.update_model(name)
        return {
            'message': msg
        }


@ns_conf.route('/<name>/adjust')
@ns_conf.param('name', 'The predictor identifier')
class PredictorAdjust(Resource):
    @ns_conf.doc('post_predictor_adjust', params=predictor_query_params)
    def post(self, name):
        data = request.json

        ds_name = data.get('data_source_name') if data.get('data_source_name') is not None else data.get('from_data')
        from_data = request.default_store.get_datasource_obj(ds_name, raw=True)

        if from_data is None:
            return {'message': f'Can not find datasource: {ds_name}'}, 400

        model_names = [x['name'] for x in request.model_interface.get_models()]
        if name not in model_names:
            return abort(404, f'Predictor "{name}" doesn\'t exist',)

        request.model_interface.adjust(
            name,
            from_data,
            request.default_store.get_datasource(ds_name)['id']
        )

        return '', 200


@ns_conf.route('/<name>/predict')
@ns_conf.param('name', 'The predictor identifier')
class PredictorPredict(Resource):
    @ns_conf.doc('post_predictor_predict', params=predictor_query_params)
    def post(self, name):
        '''Queries predictor'''
        when = request.json.get('when')

        # list is a required type for TS prediction
        if isinstance(when, (dict, list)) is False or len(when) == 0:
            return 'No data provided for the predictions', 400

        results = request.model_interface.predict(name, when, 'explain')
        return results


@ns_conf.route('/<name>/predict_datasource')
@ns_conf.param('name', 'The predictor identifier')
class PredictorPredictFromDataSource(Resource):
    @ns_conf.doc('post_predictor_predict', params=predictor_query_params)
    def post(self, name):
        data = request.json
        use_raw = False

        from_data = request.default_store.get_datasource_obj(data.get('data_source_name'), raw=use_raw)
        if from_data is None:
            abort(400, 'No valid datasource given')

        results = request.model_interface.predict(name, from_data, 'explain')
        return results


@ns_conf.route('/<name>/rename')
@ns_conf.param('name', 'The predictor identifier')
class PredictorDownload(Resource):
    @ns_conf.doc('get_predictor_download')
    def get(self, name):
        try:
            new_name = request.args.get('new_name')
            request.model_interface.rename_model(name, new_name)
        except Exception as e:
            return str(e), 400

        return f'Renamed model to {new_name}', 200


@ns_conf.route('/generate/<name>')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorGenerate(Resource):
    def put(self, name):
        problem_definition = request.json['problem_definition']
        datasource_name = request.json['data_source_name']

        from_data = request.default_store.get_datasource_obj(
            datasource_name,
            raw=True
        )
        datasource = request.default_store.get_datasource(datasource_name)

        request.model_interface.generate_predictor(
            name,
            from_data,
            datasource['id'],
            problem_definition,
            request.json.get('join_learn_process', False)
        )

        return '', 200


@ns_conf.route('/<name>/edit/json_ai')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorEditJsonAI(Resource):
    def put(self, name):
        request.model_interface.edit_json_ai(name, request.json['json_ai'])
        return '', 200


@ns_conf.route('/<name>/edit/code')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorEditCode(Resource):
    def put(self, name):
        request.model_interface.edit_code(name, request.json['code'])
        return '', 200


@ns_conf.route('/<name>/train')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorTrain(Resource):
    def put(self, name):
        for param in ['data_source_name']:
            if param not in request.json:
                return abort(400, 'Please provide {}'.format(param))

        from_data = request.default_store.get_datasource_obj(
            request.json['data_source_name'],
            raw=True
        )

        request.model_interface.fit_predictor(name, from_data, request.json.get('join_learn_process', False))
        return '', 200


@ns_conf.route('/<name>/export')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorExport(Resource):
    def get(self, name):
        payload: json = request.model_interface.export_predictor(name)
        return payload, 200


@ns_conf.route('/<name>/import')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorExport(Resource):
    def put(self, name):
        request.model_interface.import_predictor(name, request.json)



import pytds


class MSSQLConnectionChecker:
    def __init__(self, **kwargs):
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.user = kwargs.get('user')
        self.password = kwargs.get('password')

    def _get_connnection(self):
        return pytds.connect(
            user=self.user,
            password=self.password,
            dsn=self.host,
            port=self.port,
            as_dict=True,
            autocommit=True  # .commit() doesn't work
        )

    def check_connection(self):
        try:
            conn = self._get_connnection()
            conn.close()
            connected = True
        except Exception:
            connected = False

        return connected



from mindsdb.integrations.mysql_handler.mysql_handler import MySQLHandler


class MariaDBHandler(MySQLHandler):
    """
    This handler handles connection and execution of the MariaDB statements. 
    """

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)


import numpy as np
import pandas as pd
from sqlalchemy.types import (
    Integer, Float, Text
)

from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast import Insert, Identifier, Constant, CreateTable, TableColumn, DropTables

from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.utilities.log import log


class IntegrationDataNode(DataNode):
    type = 'integration'

    def __init__(self, integration_name, data_store, ds_type):
        self.integration_name = integration_name
        self.data_store = data_store
        self.ds_type = ds_type

    def get_type(self):
        return self.type

    def get_tables(self):
        return []

    def has_table(self, tableName):
        return True

    def get_table_columns(self, tableName):
        return []

    def create_table(self, table_name_parts, columns, data, is_replace=False, is_create=False):
        # is_create - create table
        # is_replace - drop table if exists
        # is_create==False and is_replace==False: just insert

        dso, _creation_info = self.data_store.create_datasource(self.integration_name, {'query': 'select 1'})
        if hasattr(dso, 'execute') is False:
            raise Exception(f"Cant create table in {self.integration_name}")
        if self.ds_type not in ('postgres', 'mysql', 'mariadb'):
            raise Exception(f'At this moment is no possible to create table in "{self.ds_type}"')

        if self.ds_type in ('mysql', 'mariadb'):
            dialect = 'mysql'
        elif self.ds_type == 'postgres':
            dialect = 'postgres'
        renderer = SqlalchemyRender(dialect)

        table_columns_meta = []
        table_columns = []
        for table in columns:
            for column in columns[table]:
                column_type = None
                for row in data:
                    column_value = row[table][column]
                    if isinstance(column_value, int):
                        column_type = Integer
                    elif isinstance(column_value, float):
                        column_type = Float
                    elif isinstance(column_value, str):
                        column_type = Text
                column_type = column_type or Text
                table_columns.append(
                    TableColumn(
                        name=column[-1],
                        type=column_type
                    )
                )
                table_columns_meta.append({
                    'table': table,
                    'name': column,
                    'type': column_type
                })

        if is_replace:
            # drop
            drop_ast = DropTables(
                tables=[Identifier(parts=table_name_parts)],
                if_exists=True
            )

            drop_query_str = renderer.get_string(drop_ast, with_failback=False)
            dso.execute(drop_query_str)

            is_create = True

        if is_create:

            create_table_ast = CreateTable(
                name=Identifier(parts=table_name_parts),
                columns=table_columns,
                is_replace=True
            )

            create_query_str = renderer.get_string(create_table_ast, with_failback=False)
            dso.execute(create_query_str)

        insert_columns = [Identifier(parts=[x['name'][-1]]) for x in table_columns_meta]
        formatted_data = []
        for row in data:
            new_row = []
            for column_meta in table_columns_meta:
                value = row[column_meta['table']][column_meta['name']]
                python_type = str
                if column_meta['type'] == Integer:
                    python_type = int
                elif column_meta['type'] == Float:
                    python_type = float

                try:
                    value = python_type(value) if value is not None else value
                except Exception:
                    pass
                new_row.append(value)
            formatted_data.append(new_row)

        insert_ast = Insert(
            table=Identifier(parts=table_name_parts),
            columns=insert_columns,
            values=formatted_data
        )

        query_str = renderer.get_string(insert_ast, with_failback=False)
        dso.execute(query_str)

    def select(self, query):
        if isinstance(query, str):
            query_str = query
        else:
            if self.ds_type in ('postgres', 'snowflake'):
                dialect = 'postgres'
            elif self.ds_type == 'mssql':
                dialect = 'mssql'
            else:
                dialect = 'mysql'
            render = SqlalchemyRender(dialect)
            try:
                query_str = render.get_string(query, with_failback=False)
            except Exception as e:
                log.error(f"Exception during query casting to '{dialect}' dialect. Query: {query}. Error: {e}")
                query_str = render.get_string(query, with_failback=True)

        dso, _creation_info = self.data_store.create_datasource(self.integration_name, {'query': query_str})
        df = dso.df
        df = df.replace({np.nan: None})
        data = df.to_dict(orient='records')
        column_names = list(dso.df.columns)

        for column_name in column_names:
            if pd.core.dtypes.common.is_datetime_or_timedelta_dtype(df[column_name]):
                pass_data = df[column_name].dt.to_pydatetime()
                for i, rec in enumerate(data):
                    rec[column_name] = pass_data[i].timestamp()

        if len(column_names) == 0:
            column_names = ['dataframe_is_empty']

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in df.dtypes.items()
        ]

        return data, columns_info



from contextlib import closing
import psycopg
from psycopg.pq import ExecStatus
from pandas import DataFrame
from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base_handler import DatabaseHandler
from mindsdb.api.mysql.mysql_proxy.mysql_proxy import RESPONSE_TYPE
from mindsdb.utilities.log import log


class PostgresHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the PostgreSQL statements. 
    """

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_args = kwargs
        del self.connection_args['database']
        self.dialect = 'postgresql'
        self.database = kwargs.get('database')

    def __connect(self):
        """
        Handles the connection to a PostgreSQL database insance.
        """
        # TODO: Check psycopg_pool
        self.connection_args['dbname'] = self.database
        connection = psycopg.connect(**self.connection_args, connect_timeout=10)
        return connection

    def check_status(self):
        """
        Check the connection of the PostgreSQL database
        :return: success status and error message if error occurs
        """
        status = {
            'success': False
        }
        try:
            con = self.__connect()
            with closing(con) as con:
                with con.cursor() as cur:
                    cur.execute('select 1;')
            status['success'] = True
        except psycopg.Error as e:
            log.error(f'Error connecting to PostgreSQL {self.database}, {e}!')
            status['error'] = e
        return status

    def native_query(self, query):
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in PostgreSQL
        :return: returns the records from the current recordset
        """
        con = self.__connect()
        with closing(con) as con:
            with con.cursor() as cur:
                try:
                    cur.execute(query)
                    if ExecStatus(cur.pgresult.status) == ExecStatus.COMMAND_OK:
                        response = {
                            'type': RESPONSE_TYPE.OK
                        }
                    else:
                        result = cur.fetchall()
                        response = {
                            'type': RESPONSE_TYPE.TABLE,
                            'data_frame': DataFrame(
                                result,
                                columns=[x.name for x in cur.description]
                            )
                        }
                except Exception as e:
                    log.error(f'Error running query: {query} on {self.database}!')
                    response = {
                        'type': RESPONSE_TYPE.ERROR,
                        'error_code': 0,
                        'error_message': str(e)
                    }
        return response

    def get_tables(self):
        """
        List all tabels in PostgreSQL without the system tables information_schema and pg_catalog
        """
        query = "SELECT * FROM information_schema.tables WHERE \
                 table_schema NOT IN ('information_schema', 'pg_catalog')"
        res = self.native_query(query)
        return res

    def get_views(self):
        """
        List all views in PostgreSQL without the system views information_schema and pg_catalog
        """
        query = "SELECT * FROM information_schema.views WHERE table_schema NOT IN ('information_schema', 'pg_catalog')"
        result = self.native_query(query)
        return result

    def describe_table(self, table_name):
        """
        List names and data types about the table coulmns
        """
        query = f"SELECT table_name, column_name, data_type FROM \
              information_schema.columns WHERE table_name='{table_name}';"
        result = self.native_query(query)
        return result

    def query(self, query):
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        renderer = SqlalchemyRender('postgres')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    # TODO: JOIN, SELECT INTO


import json
import shutil
import os
from pathlib import Path

import pandas as pd
from mindsdb_sql import parse_sql

import mindsdb_datasources
from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.interfaces.model.model_interface import ModelInterface
from mindsdb_datasources import (
    FileDS, ClickhouseDS, MariaDS, MySqlDS, PostgresDS, MSSQLDS, MongoDS,
    SnowflakeDS, AthenaDS, CassandraDS, ScyllaDS, TrinoDS
)
from mindsdb.utilities.config import Config
from mindsdb.utilities.log import log
from mindsdb.utilities.json_encoder import CustomJSONEncoder
from mindsdb.utilities.with_kwargs_wrapper import WithKWArgsWrapper
from mindsdb.interfaces.storage.db import session, Dataset, Semaphor, Predictor, Analysis, File
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.interfaces.database.integrations import IntegrationController
from mindsdb.interfaces.database.views import ViewController


class QueryDS:
    def __init__(self, query, source, source_type, company_id):
        self.query = query
        self.source = source
        self.source_type = source_type
        self.company_id = company_id

    def query(self, q):
        pass

    @property
    def df(self):
        view_interface = WithKWArgsWrapper(
            ViewController(),
            company_id=self.company_id
        )

        integration_controller = WithKWArgsWrapper(
            IntegrationController(),
            company_id=self.company_id
        )

        data_store = WithKWArgsWrapper(
            DataStore(),
            company_id=self.company_id
        )

        query = self.query
        if self.source_type == 'view_query':
            if isinstance(query, str):
                query = parse_sql(query, dialect='mysql')

            table = query.from_table.parts[-1]
            view_metadata = view_interface.get(name=table)

            integration = integration_controller.get_by_id(view_metadata['integration_id'])
            integration_name = integration['name']

            dataset_name = data_store.get_vacant_name(table)
            data_store.save_datasource(dataset_name, integration_name, {'query': view_metadata['query']})
            try:
                dataset_object = data_store.get_datasource_obj(dataset_name)
                data_df = dataset_object.df
            finally:
                data_store.delete_datasource(dataset_name)
        else:
            raise Exception(f'Unknown source_type: {self.source_type}')
        return data_df


class DataStore():
    def __init__(self):
        self.config = Config()
        self.fs_store = FsStore()
        self.dir = self.config['paths']['datasources']
        self.model_interface = ModelInterface()

    def get_analysis(self, name, company_id=None):
        dataset_record = session.query(Dataset).filter_by(company_id=company_id, name=name).first()
        if dataset_record.analysis_id is None:
            return None
        analysis_record = session.query(Analysis).get(dataset_record.analysis_id)
        if analysis_record is None:
            return None
        analysis = json.loads(analysis_record.analysis)
        return analysis

    def start_analysis(self, name, company_id=None):
        dataset_record = session.query(Dataset).filter_by(company_id=company_id, name=name).first()
        if dataset_record.analysis_id is not None:
            return None

        semaphor_record = session.query(Semaphor).filter_by(
            company_id=company_id,
            entity_id=dataset_record.id,
            entity_type='dataset'
        ).first()

        if semaphor_record is None:
            semaphor_record = Semaphor(
                company_id=company_id,
                entity_id=dataset_record.id,
                entity_type='dataset',
                action='write'
            )
            session.add(semaphor_record)
            session.commit()
        else:
            return

        try:
            analysis = self.model_interface.analyse_dataset(
                ds=self.get_datasource_obj(name, raw=True, company_id=company_id),
                company_id=company_id
            )
            dataset_record = session.query(Dataset).filter_by(company_id=company_id, name=name).first()
            analysis_record = Analysis(analysis=json.dumps(analysis, cls=CustomJSONEncoder))
            session.add(analysis_record)
            session.flush()
            dataset_record.analysis_id = analysis_record.id
            session.commit()
        except Exception as e:
            log.error(e)
        finally:
            semaphor_record = session.query(Semaphor).filter_by(company_id=company_id, entity_id=dataset_record.id, entity_type='dataset').first()
            session.delete(semaphor_record)
            session.commit()

    def get_datasets(self, name=None, company_id=None):
        dataset_arr = []
        if name is not None:
            dataset_record_arr = session.query(Dataset).filter_by(company_id=company_id, name=name)
        else:
            dataset_record_arr = session.query(Dataset).filter_by(company_id=company_id)
        for dataset_record in dataset_record_arr:
            try:
                if dataset_record.data is None:
                    continue
                dataset = json.loads(dataset_record.data)
                dataset['created_at'] = dataset_record.created_at
                dataset['updated_at'] = dataset_record.updated_at
                dataset['name'] = dataset_record.name
                dataset['id'] = dataset_record.id
                dataset_arr.append(dataset)
            except Exception as e:
                log.error(e)
        return dataset_arr

    def get_files_names(self, company_id=None):
        """ return list of files names
        """
        return [x[0] for x in session.query(File.name).filter_by(company_id=company_id)]

    def get_file_meta(self, name, company_id=None):
        file_record = session.query(File).filter_by(company_id=company_id, name=name).first()
        if file_record is None:
            return None
        columns = file_record.columns
        if isinstance(columns, str):
            columns = json.loads(columns)
        return {
            'name': file_record.name,
            'columns': columns,
            'row_count': file_record.row_count
        }

    def get_data(self, name, where=None, limit=None, offset=None, company_id=None):
        offset = 0 if offset is None else offset
        ds = self.get_datasource_obj(name, company_id=company_id)

        if limit is not None:
            # @TODO Add `offset` to the `filter` method of the datasource and get rid of `offset`
            filtered_ds = ds.filter(where=where, limit=limit + offset).iloc[offset:]
        else:
            filtered_ds = ds.filter(where=where)

        filtered_ds = filtered_ds.where(pd.notnull(filtered_ds), None)
        data = filtered_ds.to_dict(orient='records')
        return {
            'data': data,
            'rowcount': len(ds),
            'columns_names': list(data[0].keys())
        }

    def get_datasource(self, name, company_id=None):
        dataset_arr = self.get_datasets(name, company_id=company_id)
        if len(dataset_arr) == 1:
            return dataset_arr[0]
        # @TODO: Remove when db swithc is more stable, this should never happen, but good santiy check while this is kinda buggy
        elif len(dataset_arr) > 1:
            log.error('Two or more dataset with the same name, (', len(dataset_arr), ') | Full list: ', dataset_arr)
            raise Exception('Two or more dataset with the same name')
        return None

    def delete_datasource(self, name, company_id=None):
        dataset_record = Dataset.query.filter_by(company_id=company_id, name=name).first()
        if not Config()["force_dataset_removing"]:
            linked_models = Predictor.query.filter_by(company_id=company_id, dataset_id=dataset_record.id).all()
            if linked_models:
                raise Exception("Can't delete {} dataset because there are next models linked to it: {}".format(name, [model.name for model in linked_models]))
        session.query(Semaphor).filter_by(
            company_id=company_id, entity_id=dataset_record.id, entity_type='dataset'
        ).delete()
        session.delete(dataset_record)
        session.commit()
        self.fs_store.delete(f'datasource_{company_id}_{dataset_record.id}')  # TODO del in future
        try:
            shutil.rmtree(os.path.join(self.dir, f'{company_id}@@@@@{name}'))
        except Exception:
            pass

    def get_vacant_name(self, base=None, company_id=None):
        ''' returns name of dataset, which starts from 'base' and ds with that name is not exists yet
        '''
        if base is None:
            base = 'dataset'
        datasets = session.query(Dataset.name).filter_by(company_id=company_id).all()
        datasets_names = [x[0] for x in datasets]
        if base not in datasets_names:
            return base
        for i in range(1, 1000):
            candidate = f'{base}_{i}'
            if candidate not in datasets_names:
                return candidate
        raise Exception(f"Can not find appropriate name for dataset '{base}'")

    def create_datasource(self, source_type, source, file_path=None, company_id=None):
        integration_controller = IntegrationController()
        if source_type == 'view_query':
            dsClass = QueryDS
            creation_info = {
                'class': dsClass.__name__,
                'args': [],
                'kwargs': {
                    'query': source['query'],
                    'source': source['source'],   # view
                    'source_type': source_type,
                    'company_id': company_id
                }
            }

            ds = dsClass(**creation_info['kwargs'])
        elif source_type == 'file':
            file_name = source.get('mindsdb_file_name')
            file_record = session.query(File).filter_by(company_id=company_id, name=file_name).first()
            if file_record is None:
                raise Exception(f"Cant find file '{file_name}'")
            remote_file_path = f'file_{company_id}_{file_record.id}'
            self.fs_store.get(remote_file_path, remote_file_path, self.dir)
            kwargs = {}
            query = source.get('query')
            if query is not None:
                kwargs['query'] = query

            path = Path(self.dir).joinpath(remote_file_path).joinpath(file_record.source_file_path)

            creation_info = {
                'class': 'FileDS',
                'args': [str(path)],
                'kwargs': kwargs
            }
            ds = FileDS(str(path), **kwargs)

        elif integration_controller.get(source_type, company_id) is not None:
            integration = integration_controller.get(source_type, company_id)

            ds_class_map = {
                'clickhouse': ClickhouseDS,
                'mariadb': MariaDS,
                'mysql': MySqlDS,
                'singlestore': MySqlDS,
                'postgres': PostgresDS,
                'cockroachdb': PostgresDS,
                'mssql': MSSQLDS,
                'mongodb': MongoDS,
                'snowflake': SnowflakeDS,
                'athena': AthenaDS,
                'cassandra': CassandraDS,
                'scylladb': ScyllaDS,
                'trinodb': TrinoDS,
                'questdb': PostgresDS
            }

            try:
                dsClass = ds_class_map[integration['type']]
            except KeyError:
                raise KeyError(f"Unknown DS type: {source_type}, type is {integration['type']}")

            if dsClass is None:
                raise Exception(f"Unsupported dataset: {source_type}, type is {integration['type']}, please install required dependencies!")

            if integration['type'] in ['clickhouse']:
                creation_info = {
                    'class': dsClass.__name__,
                    'args': [],
                    'kwargs': {
                        'query': source['query'],
                        'user': integration['user'],
                        'password': integration['password'],
                        'host': integration['host'],
                        'port': integration['port']
                    }
                }
                ds = dsClass(**creation_info['kwargs'])

            elif integration['type'] in ['mssql', 'postgres', 'cockroachdb', 'mariadb', 'mysql', 'singlestore', 'cassandra', 'scylladb', 'questdb']:
                creation_info = {
                    'class': dsClass.__name__,
                    'args': [],
                    'kwargs': {
                        'query': source['query'],
                        'user': integration['user'],
                        'password': integration['password'],
                        'host': integration['host'],
                        'port': integration['port']
                    }
                }
                kwargs = creation_info['kwargs']

                integration_folder_name = f'integration_files_{company_id}_{integration["id"]}'
                if integration['type'] in ('mysql', 'mariadb'):
                    kwargs['ssl'] = integration.get('ssl')
                    kwargs['ssl_ca'] = integration.get('ssl_ca')
                    kwargs['ssl_cert'] = integration.get('ssl_cert')
                    kwargs['ssl_key'] = integration.get('ssl_key')
                    for key in ['ssl_ca', 'ssl_cert', 'ssl_key']:
                        if isinstance(kwargs[key], str) and len(kwargs[key]) > 0:
                            kwargs[key] = os.path.join(
                                self.integrations_dir,
                                integration_folder_name,
                                kwargs[key]
                            )
                elif integration['type'] in ('cassandra', 'scylla'):
                    kwargs['secure_connect_bundle'] = integration.get('secure_connect_bundle')
                    if (
                        isinstance(kwargs['secure_connect_bundle'], str)
                        and len(kwargs['secure_connect_bundle']) > 0
                    ):
                        kwargs['secure_connect_bundle'] = os.path.join(
                            self.integrations_dir,
                            integration_folder_name,
                            kwargs['secure_connect_bundle']
                        )

                if 'database' in integration:
                    kwargs['database'] = integration['database']

                if 'database' in source:
                    kwargs['database'] = source['database']

                ds = dsClass(**kwargs)

            elif integration['type'] == 'snowflake':
                creation_info = {
                    'class': dsClass.__name__,
                    'args': [],
                    'kwargs': {
                        'query': source['query'],
                        'schema': source.get('schema', integration['schema']),
                        'warehouse': source.get('warehouse', integration['warehouse']),
                        'database': source.get('database', integration['database']),
                        'host': integration['host'],
                        'password': integration['password'],
                        'user': integration['user'],
                        'account': integration['account']
                    }
                }

                ds = dsClass(**creation_info['kwargs'])

            elif integration['type'] == 'mongodb':
                if isinstance(source['find'], str):
                    source['find'] = json.loads(source['find'])
                creation_info = {
                    'class': dsClass.__name__,
                    'args': [],
                    'kwargs': {
                        'database': source['database'],
                        'collection': source['collection'],
                        'query': source['find'],
                        'user': integration['user'],
                        'password': integration['password'],
                        'host': integration['host'],
                        'port': integration['port']
                    }
                }

                ds = dsClass(**creation_info['kwargs'])

            elif integration['type'] == 'athena':
                creation_info = {
                    'class': dsClass.__name__,
                    'args': [],
                    'kwargs': {
                        'query': source['query'],
                        'staging_dir': source['staging_dir'],
                        'database': source['database'],
                        'access_key': source['access_key'],
                        'secret_key': source['secret_key'],
                        'region_name': source['region_name']
                    }
                }

                ds = dsClass(**creation_info['kwargs'])

            elif integration['type'] == 'trinodb':
                creation_info = {
                    'class': dsClass.__name__,
                    'args': [],
                    'kwargs': {
                        'query': source['query'],
                        'user': integration['user'],
                        'password': integration['password'],
                        'host': integration['host'],
                        'port': integration['port'],
                        'schema': integration['schema'],
                        'catalog': integration['catalog']
                    }
                }

                ds = dsClass(**creation_info['kwargs'])
        else:
            # This probably only happens for urls
            ds = FileDS(source)
            creation_info = {
                'class': 'FileDS',
                'args': [source],
                'kwargs': {}
            }
        return ds, creation_info

    def get_files(self, company_id=None):
        """ Get list of files

            Returns:
                list[dict]: files metadata
        """
        file_records = session.query(File).filter_by(company_id=company_id).all()
        files_metadata = [{
            'name': record.name,
            'row_count': record.row_count,
            'columns': record.columns,
        } for record in file_records]
        return files_metadata

    def save_file(self, name, file_path, file_name=None, company_id=None):
        """ Save the file to our store

            Args:
                name (str): with that name file will be available in sql api
                file_name (str): file name
                file_path (str): path to the file
                company_id (int): company id

            Returns:
                int: id of 'file' record in db
        """
        files_metadata = self.get_files()
        if name in [x['name'] for x in files_metadata]:
            raise Exception(f'File already exists: {name}')

        if file_name is None:
            file_name = Path(file_path).name

        try:
            ds = FileDS(str(file_path))
            ds_meta = self._get_ds_meta(ds)

            column_names = ds_meta['column_names']
            if ds_meta['column_names'] is not None:
                column_names = [dict(name=x) for x in ds_meta['column_names']]
            file_record = File(
                name=name,
                company_id=company_id,
                source_file_path=file_name,
                file_path='',
                row_count=ds_meta['row_count'],
                columns=column_names
            )
            session.add(file_record)
            session.commit()
            store_file_path = f'file_{company_id}_{file_record.id}'
            file_record.file_path = store_file_path
            session.commit()

            file_dir = Path(self.dir).joinpath(store_file_path)
            file_dir.mkdir(parents=True, exist_ok=True)
            source = file_dir.joinpath(file_name)
            # NOTE may be delay between db record exists and file is really in folder
            shutil.move(file_path, str(source))

            self.fs_store.put(store_file_path, store_file_path, self.dir)
        except Exception as e:
            log.error(e)
            raise
        finally:
            shutil.rmtree(file_dir)

        return file_record.id

    def delete_file(self, name, company_id):
        file_record = session.query(File).filter_by(company_id=company_id, name=name).first()
        if file_record is None:
            return None
        file_id = file_record.id
        session.delete(file_record)
        session.commit()
        self.fs_store.delete(f'file_{company_id}_{file_id}')
        return True

    def _get_ds_meta(self, ds):
        if hasattr(ds, 'get_columns') and hasattr(ds, 'get_row_count'):
            try:
                column_names = ds.get_columns()
                row_count = ds.get_row_count()
            except Exception:
                df = ds.df
                column_names = list(df.keys())
                row_count = len(df)
        else:
            df = ds.df
            column_names = list(df.keys())
            row_count = len(df)
        return {
            'column_names': column_names,
            'row_count': row_count
        }

    def save_datasource(self, name, source_type, source=None, file_path=None, company_id=None):
        dataset_record = session.query(Dataset).filter_by(company_id=company_id, name=name).first()
        while dataset_record is not None:
            raise Exception(f'Dataset with name {name} already exists')

        if source_type == 'views':
            source_type = 'view_query'
        elif source_type == 'files':
            source_type = 'file'

        try:
            dataset_record = Dataset(
                company_id=company_id,
                name=name,
                datasources_version=mindsdb_datasources.__version__,
                mindsdb_version=mindsdb_version
            )
            session.add(dataset_record)
            session.commit()

            ds, creation_info = self.create_datasource(source_type, source, file_path, company_id)

            ds_meta = self._get_ds_meta(ds)
            column_names = ds_meta['column_names']
            row_count = ds_meta['row_count']

            dataset_record.ds_class = creation_info['class']
            dataset_record.creation_info = json.dumps(creation_info)
            dataset_record.data = json.dumps({
                'source_type': source_type,
                'source': source,
                'row_count': row_count,
                'columns': [dict(name=x) for x in column_names]
            })

            session.commit()

        except Exception as e:
            log.error(f'Error creating dataset {name}, exception: {e}')
            try:
                self.delete_datasource(name, company_id=company_id)
            except Exception:
                pass
            raise e

        return self.get_datasource_obj(name, raw=True, company_id=company_id)

    def get_datasource_obj(self, name=None, id=None, raw=False, company_id=None):
        try:
            if name is not None:
                dataset_record = session.query(Dataset).filter_by(company_id=company_id, name=name).first()
            else:
                dataset_record = session.query(Dataset).filter_by(company_id=company_id, id=id).first()

            creation_info = json.loads(dataset_record.creation_info)
            if raw:
                return creation_info
            else:
                if dataset_record.ds_class == 'FileDS':
                    file_record = session.query(File).filter_by(company_id=company_id, name=name).first()
                    if file_record is not None:
                        self.fs_store.get(f'{company_id}@@@@@{dataset_record.name}', f'file_{company_id}_{file_record.id}', self.dir)
                return eval(creation_info['class'])(*creation_info['args'], **creation_info['kwargs'])
        except Exception as e:
            log.error(f'Error getting dataset {name}, exception: {e}')
            return None

    def get_file_path(self, name, company_id):
        file_record = session.query(File).filter_by(company_id=company_id, name=name).first()
        if file_record is None:
            raise Exception("File '{name}' does not exists")
        file_dir = f'file_{company_id}_{file_record.id}'
        self.fs_store.get(file_dir, file_dir, self.dir)
        return str(Path(self.dir).joinpath(file_dir).joinpath(Path(file_record.source_file_path).name))



from mindsdb_datasources import SnowflakeDS


class SnowflakeConnectionChecker:
    def __init__(self, **kwargs):
        self.host = kwargs.get('host')
        self.user = kwargs.get('user')
        self.password = kwargs.get('password')
        self.account = kwargs.get('account')
        self.warehouse = kwargs.get('warehouse')
        self.database = kwargs.get('database')
        self.schema = kwargs.get('schema')
        self.protocol = kwargs.get('protocol')
        self.port = kwargs.get('port')

    def check_connection(self):
        connected = False
        try:
            ds = SnowflakeDS('SELECT 1;', self.host, self.user, self.password, self.account, self.warehouse, self.database, self.schema, self.protocol, self.port)
            assert len(ds.df) == 1
            connected = True
        except Exception as e:
            print(e)
            connected = False
        return connected



from bson.int64 import Int64
from collections import OrderedDict
from lightwood.api import dtype
from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'find': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        models = mindsdb_env['model_interface'].get_models()
        model_names = [x['name'] for x in models]
        table = query['find']
        where_data = query.get('filter', {})
        if table == 'predictors':
            data = [{
                'name': x['name'],
                'status': x['status'],
                'accuracy': str(x['accuracy']) if x['accuracy'] is not None else None,
                'predict': ', '.join(x['predict'] if isinstance(x['predict'], list) else [x['predict']]),
                'error': x['error'],
                'select_data_query': '',
                'training_options': ''
            } for x in models]
        elif table in model_names:
            # prediction
            model = mindsdb_env['model_interface'].get_model_data(name=query['find'])

            columns = []
            columns += list(model['dtype_dict'].keys())
            predict = model['predict']
            if not isinstance(predict, list):
                predict = [predict]
            columns += [f'{x}_original' for x in predict]
            for col in predict:
                if model['dtype_dict'][col] in (dtype.integer, dtype.float):
                    columns += [f"{col}_min", f"{col}_max"]
                columns += [f"{col}_confidence"]
                columns += [f"{col}_explain"]

            columns += ['when_data', 'select_data_query']

            where_data_list = where_data if isinstance(where_data, list) else [where_data]
            for statement in where_data_list:
                if isinstance(statement, dict):
                    for key in statement:
                        if key not in columns:
                            columns.append(key)

            datasource = where_data
            if 'select_data_query' in where_data:
                integrations = mindsdb_env['datasource_controller'].get_all().keys()
                connection = where_data.get('connection')
                if connection is None:
                    if 'default_mongodb' in integrations:
                        connection = 'default_mongodb'
                    else:
                        for integration in integrations:
                            if integration.startswith('mongodb_'):
                                connection = integration
                                break

                if connection is None:
                    raise Exception("Can't find connection from which fetch data")

                ds_name = mindsdb_env['data_store'].get_vacant_name('temp')

                mindsdb_env['data_store'].save_datasource(
                    name=ds_name,
                    source_type=connection,
                    source=where_data['select_data_query']
                )
                datasource = mindsdb_env['data_store'].get_datasource_obj(ds_name, raw=True)

            if isinstance(datasource, OrderedDict):
                datasource = dict(datasource)

            pred_dict_arr, explanations = mindsdb_env['model_interface'].predict(table, datasource, 'dict&explain')

            if 'select_data_query' in where_data:
                mindsdb_env['data_store'].delete_datasource(ds_name)

            predicted_columns = model['predict']
            if not isinstance(predicted_columns, list):
                predicted_columns = [predicted_columns]

            data = []
            all_columns = list(model['dtype_dict'].keys())   # [k for k in pred_dict_arr[0] if k in columns]
            min_max_keys = []
            for col in predicted_columns:
                if model['dtype_dict'][col] in (dtype.integer, dtype.float):
                    min_max_keys.append(col)

            for i in range(len(pred_dict_arr)):
                row = {}
                explanation = explanations[i]

                for value in pred_dict_arr[i].values():
                    row.update(value)
                if 'predicted_value' in row:
                    del row['predicted_value']
                for key in pred_dict_arr[i]:
                    row[key] = pred_dict_arr[i][key]['predicted_value']
                for key in all_columns:
                    if key not in row:
                        row[key] = None

                for key in predicted_columns:
                    row[key + '_confidence'] = explanation[key]['confidence']
                    row[key + '_explain'] = explanation[key]
                for key in min_max_keys:
                    if 'confidence_lower_bound' in explanation[key]:
                        row[key + '_min'] = explanation[key]['confidence_lower_bound']
                    if 'confidence_upper_bound' in explanation[key]:
                        row[key + '_max'] = explanation[key]['confidence_upper_bound']
                data.append(row)

        else:
            # probably wrong table name. Mongo in this case returns empty data
            data = []

        if 'projection' in query and len(data) > 0:
            true_filter = []
            false_filter = []
            for key, value in query['projection'].items():
                if helpers.is_true(value):
                    true_filter.append(key)
                else:
                    false_filter.append(key)

            keys = list(data[0].keys())
            del_id = '_id' in false_filter
            if len(true_filter) > 0:
                for row in data:
                    for key in keys:
                        if key != '_id':
                            if key not in true_filter:
                                del row[key]
                        elif del_id:
                            del row[key]
            else:
                for row in data:
                    for key in false_filter:
                        if key in row:
                            del row[key]

        db = mindsdb_env['config']['api']['mongodb']['database']

        cursor = {
            'id': Int64(0),
            'ns': f"{db}.$cmd.{query['find']}",
            'firstBatch': data
        }
        return {
            'cursor': cursor,
            'ok': 1
        }


responder = Responce()



from mindsdb_datasources import TrinoDS


class TrinodbConnectionChecker:
    def __init__(self, **kwargs):
        self.user = kwargs.get('user')
        self.password = kwargs.get('password')
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.catalog = kwargs.get('catalog')
        self.schema = kwargs.get('schema')

    def check_connection(self):
        try:
            ds = TrinoDS('SELECT 1', self.user, self.password,  self.host, self.port,  self.catalog, self.schema )
            assert len(ds.df) == 1
        except Exception as e:
            return False
        else:
            return True



from mindsdb.api.mongo.classes import Responder
from mindsdb.interfaces.storage.db import session, Dataset
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'insert': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            res = self._result(query, request_env, mindsdb_env)
        except Exception as e:
            res = {
                'n': 0,
                'writeErrors': [{
                    'index': 0,
                    'code': 0,
                    'errmsg': str(e)
                }],
                'ok': 1
            }
        return res

    def _result(self, query, request_env, mindsdb_env):
        table = query['insert']
        if table != 'predictors':
            raise Exception("Only insert to 'predictors' table allowed")

        predictors_columns = [
            'name',
            'status',
            'accuracy',
            'predict',
            'select_data_query',
            'training_options',
            'connection'
        ]

        models = mindsdb_env['model_interface'].get_models()

        if len(query['documents']) != 1:
            raise Exception("Must be inserted just one predictor at time")

        for doc in query['documents']:
            if '_id' in doc:
                del doc['_id']

            bad_columns = [x for x in doc if x not in predictors_columns]
            if len(bad_columns) > 0:
                raise Exception(f"Is no possible insert this columns to 'predictors' collection: {', '.join(bad_columns)}")

            if 'name' not in doc:
                raise Exception("Please, specify 'name' field")

            if 'predict' not in doc:
                raise Exception("Please, specify 'predict' field")

            if doc['name'] in [x['name'] for x in models]:
                raise Exception(f"Predictor with name '{doc['name']}' already exists")

            select_data_query = doc.get('select_data_query')
            if select_data_query is None:
                raise Exception("'select_data_query' must be in query")

            kwargs = doc.get('training_options', {})

            integrations = mindsdb_env['datasource_controller'].get_all().keys()
            connection = doc.get('connection')
            if connection is None:
                if 'default_mongodb' in integrations:
                    connection = 'default_mongodb'
                else:
                    for integration in integrations:
                        if integration.startswith('mongodb_'):
                            connection = integration
                            break

            if connection is None:
                raise Exception("Can't find connection for data source")

            ds_name = mindsdb_env['data_store'].get_vacant_name(doc['name'])

            select_data_query = select_data_query if isinstance(select_data_query, dict) else {'query': select_data_query}
            ds = mindsdb_env['data_store'].save_datasource(
                name=ds_name,
                source_type=connection,
                source=select_data_query
            )

            predict = doc['predict']
            if not isinstance(predict, list):
                predict = [x.strip() for x in predict.split(',')]

            ds_columns = [x['name'] for x in mindsdb_env['data_store'].get_datasource(ds_name)['columns']]
            for col in predict:
                if col not in ds_columns:
                    mindsdb_env['data_store'].delete_datasource(ds_name)
                    raise Exception(f"Column '{col}' not exists")

            datasource_record = session.query(Dataset).filter_by(company_id=mindsdb_env['company_id'], name=ds_name).first()
            mindsdb_env['model_interface'].learn(
                doc['name'],
                ds,
                predict,
                datasource_record.id,
                kwargs=dict(kwargs),
                delete_ds_on_fail=True,
                user_class=mindsdb_env.get('user_class', 0)
            )

        result = {
            "n": len(query['documents']),
            "ok": 1
        }

        return result


responder = Responce()



from pandas import DataFrame as DF
from mindsdb_sql import parse_sql

from mindsdb.api.mysql.mysql_proxy.classes.sql_query import get_all_tables
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df


class ViewDataNode(DataNode):
    type = 'view'

    def __init__(self, view_interface, integration_controller, data_store):
        self.view_interface = view_interface
        self.integration_controller = integration_controller
        self.data_store = data_store

    def get_tables(self):
        views = self.view_interface.get_all()
        return list(views.keys())

    def has_table(self, table):
        views = self.view_interface.get_all()
        return table in views

    def get_table_columns(self, table):
        # TODO
        raise Exception('not iomplemented')

    def select(self, query):
        if isinstance(query, str):
            query = parse_sql(query, dialect='mysql')
        query_str = str(query)

        table = query.from_table.parts[-1]
        view_metadata = self.view_interface.get(name=table)

        integration = self.integration_controller.get_by_id(view_metadata['integration_id'])
        integration_name = integration['name']

        dataset_name = self.data_store.get_vacant_name(table)
        self.data_store.save_datasource(dataset_name, integration_name, {'query': view_metadata['query']})
        try:
            dataset_object = self.data_store.get_datasource_obj(dataset_name)
            data_df = dataset_object.df
        finally:
            self.data_store.delete_datasource(dataset_name)

        result = query_df(data_df, query_str)

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in result.dtypes.items()
        ]

        return result.to_dict(orient='records'), columns_info



from mindsdb.interfaces.storage.db import session, Integration, View


class ViewController:
    def add(self, name, query, integration_name, company_id=None):
        integration_records = session.query(Integration).filter_by(company_id=company_id).all()
        integration_id = None
        for record in integration_records:
            if record.name == integration_name:
                integration_id = record.id
                break
        else:
            raise Exception(f"Can't find integration with name: {integration_name}")

        view_record = View(name=name, company_id=company_id, query=query, integration_id=integration_id)
        session.add(view_record)
        session.commit()

    def _get_view_record_data(self, record):
        return {
            'name': record.name,
            'query': record.query,
            'integration_id': record.integration_id
        }

    def get(self, id=None, name=None, company_id=None):
        if id is not None:
            records = session.query(View).filter_by(id=id, company_id=company_id).all()
        elif name is not None:
            records = session.query(View).filter_by(name=name, company_id=company_id).all()
        if len(records) == 0:
            raise Exception(f"Can't find view with name/id: {name}/{id}")
        elif len(records) > 1:
            raise Exception(f"There are multiple views with name/id: {name}/{id}")
        record = records[0]
        return self._get_view_record_data(record)

    def get_all(self, company_id=None):
        view_records = session.query(View).filter_by(company_id=company_id).all()
        views_dict = {}
        for record in view_records:
            views_dict[record.name] = self._get_view_record_data(record)
        return views_dict



from mindsdb.integrations.libs.base_handler import DatabaseHandler
from mindsdb_sql import parse_sql
import pymssql
from mindsdb.utilities.log import log
from contextlib import closing
from mindsdb.api.mysql.mysql_proxy.mysql_proxy import RESPONSE_TYPE
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
import pandas as pd

class SqlServerHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Microsoft SQL Server statements. 
    """

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_args = kwargs
        self.dialect = 'mssql'
        self.database = kwargs.get('database')

    def __connect(self):
        """
        Handles the connection to a SQL Server insance.
        """
        connection = pymssql.connect(**self.connection_args)
        return connection

    def check_status(self):
        """
        Check the connection of the SQL Server database
        :return: success status and error message if error occurs
        """
        status = {
            'success': False
        }
        try:
            con = self.__connect()
            with closing(con) as con:
                #TODO: best way to check con.connected ?             
                status['success'] = True
        except Exception as e:
            log.error(f'Error connecting to SQL Server {self.database}, {e}!')
            status['error'] = e
        return status

    def native_query(self, query):
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in SQL Server
        :return: returns the records from the current recordset
        """
        con = self.__connect()
        with closing(con) as con:
            with con.cursor(as_dict=True) as cur:
                try:
                    res = cur.execute(query)
                    result = cur.fetchall()  
                    if result:                                  
                        response = {
                            'type': RESPONSE_TYPE.TABLE,
                            'data_frame': pd.DataFrame(
                                result,
                                columns=[x[0] for x in cur.description]
                            )
                        }
                    else:
                        response = {
                            'type': RESPONSE_TYPE.OK
                        }
                except Exception as e:
                    log.error(f'Error running query: {query} on {self.database}!')
                    response = {
                        'type': RESPONSE_TYPE.ERROR,
                        'error_code': 0,
                        'error_message': str(e)
                    }
        return response

    def get_tables(self):
        """
        Get a list with all of the tabels in MySQL
        """
        q = f"SELECT * FROM {self.database}.INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE = 'BASE TABLE';"
        result = self.native_query(q)
        return result

    def get_views(self):
        """
        Get more information about specific database views
        """
        #TODO: check what info we need for views
        q = f"SELECT * FROM sys.views;"
        result = self.native_query(q)
        return result

    def describe_table(self, table_name):
        """
        Show details about the table
        """
        q = f"SELECT * FROM information_schema.columns WHERE table_name = '{table_name}';"
        result = self.native_query(q)
        return result


    def query(self, query):
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender('mysql')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)


from typing import List, Optional
from contextlib import closing

import pandas as pd
import mysql.connector
from sqlalchemy import create_engine

from mindsdb_sql import parse_sql
from mindsdb.integrations.libs.base_handler import DatabaseHandler
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.utilities.log import log
from mindsdb.api.mysql.mysql_proxy.mysql_proxy import RESPONSE_TYPE


class MySQLHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the MySQL statements. 
    """

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.mysql_url = None
        self.parser = parse_sql
        self.dialect = 'mysql'
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.user = kwargs.get('user')
        self.database = kwargs.get('database')  # todo: may want a method to change active DB
        self.password = kwargs.get('password')
        self.ssl = kwargs.get('ssl')
        self.ssl_ca = kwargs.get('ssl_ca')
        self.ssl_cert = kwargs.get('ssl_cert')
        self.ssl_key = kwargs.get('ssl_key')

    def __connect(self):
        config = {
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password
        }
        if self.ssl is True:
            config['client_flags'] = [mysql.connector.constants.ClientFlag.SSL]
            if self.ssl_ca is not None:
                config["ssl_ca"] = self.ssl_ca
            if self.ssl_cert is not None:
                config["ssl_cert"] = self.ssl_cert
            if self.ssl_key is not None:
                config["ssl_key"] = self.ssl_key

        connection = mysql.connector.connect(**config)
        return connection

    def check_status(self):
        """
        Check the connection of the MySQL database
        :return: success status and error message if error occurs
        """
        status = {
            'success': False
        }
        try:
            con = self.__connect()
            with closing(con) as con:
                status['success'] = con.is_connected()
        except Exception as e:
            log.error(f'Error connecting to MySQL {self.database}, {e}!')
            status['error'] = e
        return status

    def native_query(self, query):
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in MySQL
        :return: returns the records from the current recordset
        """
        con = self.__connect()
        with closing(con) as con:
            with con.cursor(dictionary=True, buffered=True) as cur:
                try:
                    cur.execute(f"USE {self.database};")
                    cur.execute(query)       
                    if cur.with_rows:                               
                        result = cur.fetchall()                
                        response = {
                            'type': RESPONSE_TYPE.TABLE,
                            'data_frame': pd.DataFrame(
                                result,
                                columns=[x[0] for x in cur.description]
                            )
                        }
                    else:
                        response = {
                            'type': RESPONSE_TYPE.OK
                        }
                except Exception as e:
                    log.error(f'Error running query: {query} on {self.database}!')
                    response = {
                        'type': RESPONSE_TYPE.ERROR,
                        'error_code': 0,
                        'error_message': str(e)
                    }
        return response

    def get_tables(self):
        """
        Get a list with all of the tabels in MySQL
        """
        q = "SHOW TABLES;"
        result = self.native_query(q)
        return result

    def get_views(self):
        """
        Get more information about specific database views
        """
        q = f"SHOW FULL TABLES IN {self.database} WHERE TABLE_TYPE LIKE 'VIEW';"
        result = self.native_query(q)
        return result

    def describe_table(self, table_name):
        """
        Show details about the table
        """
        q = f"DESCRIBE {table_name};"
        result = self.native_query(q)
        return result

    def select_query(self, query):
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender('mysql')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)
  
    def select_into(self, table, dataframe: pd.DataFrame):
        """
        TODO: Update this
        """
        try:
            con = create_engine(f'mysql://{self.host}:{self.port}/{self.database}', echo=False)
            dataframe.to_sql(table, con=con, if_exists='append', index=False)
            return True
        except Exception as e:
            print(e)
            raise Exception(f"Could not select into table {table}, aborting.")

    def join(self, stmt, data_handler, into: Optional[str] = None) -> pd.DataFrame:
        """
        TODO: Update this
        """
        local_result = self.select_query(stmt.targets, stmt.from_table.left, stmt.where)  # should check it's actually on the left
        external_result = data_handler.select_query(stmt.targets, stmt.from_table.right, stmt.where)  # should check it's actually on the right

        local_df = pd.DataFrame.from_records(local_result)
        external_df = pd.DataFrame.from_records(external_result)
        df = local_df.join(external_df, on=[str(t) for t in stmt.targets], lsuffix='_left', rsuffix='_right')

        if into:
            self.select_into(into, df)
        return df
